/*********************************************************************************
 *   Kenya - an introductory programming language and development environment    *
 *   Copyright (C) 2001 Robert Chatley                                           *
 *                                                                               *
 *   This program is free software; you can redistribute it and/or               *
 *   modify it under the terms of the GNU General Public License                 *
 *   as published by the Free Software Foundation; either version 2              *
 *   of the License, or (at your option) any later version.                      *
 *                                                                               *
 *   This program is distributed in the hope that it will be useful,             *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of              *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the               *
 *   GNU General Public License for more details.                                *
 *                                                                               *
 *   You should have received a copy of the GNU General Public License           *
 *   along with this program; if not, write to the Free Software                 *
 *   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA. *
 *                                                                               *
 *   The author can be contacted by email at robert@chatley.com                  *
 *                                                                               *
 *********************************************************************************/

import minijava.analysis.*;
import minijava.node.*;

import java.util.*;
import java.io.*;

public class TypeChecker extends DepthFirstAdapter {
  private boolean nowindowmode;

  KenyaUI ui;

  SymbolTable symbolTable;
  SymbolTable functions;
  HashMap knownTypes;
  boolean checked;
  boolean genCode;
  String expected;
  String found;
  String currentfunc;
  String currentclass;
  boolean expectingArrayIndex;
  int arraydim;
  int expdim, founddim;
  int curdepth , maxdepth;
  int ln;

  public TypeChecker(KenyaUI u, HashMap kt, SymbolTable s)
  {
    this();
    ui = u;
    symbolTable = s;
    knownTypes = kt;
    functions = new SymbolTable();

    addStandardTypes();
  }

  /**
   * 
   * @param u
   * @param kt
   */
  public TypeChecker(KenyaUI u, HashMap kt)
  {
    this();
    ui = u;
    symbolTable = new SymbolTable();
    knownTypes = kt;
    functions = new SymbolTable();

    addStandardTypes();
  }

  public TypeChecker(KenyaUI u)
  {
    this();
    ui = u;
    symbolTable = new SymbolTable();
    functions = new SymbolTable();
    knownTypes = new HashMap();

    addStandardTypes();
  }

  public TypeChecker()
  {
    currentfunc = "";
    currentclass = "";
    expected = "";
    found = "";
    checked = true;
    genCode = true;
    nowindowmode = false;
    founddim = expdim = arraydim = 0;
    curdepth = maxdepth = 0;
    expectingArrayIndex = true;
  }

  /**
   * Hewewqewq
   */
  public void setNoWindows() {
    nowindowmode = true;

  }

  private void addStandardTypes() {


    knownTypes.put("void", new SymbolTable( "void" ));
    knownTypes.put("char", new SymbolTable( "char" ));
    knownTypes.put("int",  new SymbolTable( "int" ));
    knownTypes.put("double", new SymbolTable( "double" ));
    knownTypes.put("String", new SymbolTable( "String" ));
    knownTypes.put("boolean", new SymbolTable( "boolean" ));
    knownTypes.put("Vector",  new SymbolTable( "Vector" ));
    knownTypes.put("Stack", new SymbolTable( "Stack" ));
    knownTypes.put("Dictionary", new SymbolTable( "Dictionary" ));
    knownTypes.put("Hashtable", new SymbolTable( "Hashtable" ));
    knownTypes.put("Enumeration", new SymbolTable( "Enumeration" ));  
  }

  public SymbolTable getSymbolTable()
  {
    return symbolTable;
  }

  // cla97 >>
  public void caseAListStatements(AListStatements node)
  {
    expected = "";
    super.caseAListStatements(node);
  }
  // cla97 <<

  public void caseAClassDecDeclaration( AClassDecDeclaration node )
  {
    if ( node.getIdentifier() != null ) {
      String classname = node.getIdentifier().toString().trim();

      if ( knownTypes.containsKey( classname ) && knownTypes.get( classname ) != null ) {
        ln = node.getKlass().getLine();
        giveError("A type with the name " + classname + " has already been defined." , ln );
        return;
      } else {

        knownTypes.put( classname , null );

        TypeChecker ltc = new TypeChecker(ui, knownTypes);
        ltc.nowindowmode = nowindowmode;
        ltc.currentclass = classname;

        LinkedList decs = node.getDeclaration();
        ListIterator decsIterator = decs.listIterator(0);

        while ( decsIterator.hasNext() ) {
          Node t = (Node)decsIterator.next();
          t.apply(ltc);
        }

        SymbolTable s = ltc.getSymbolTable();

        if ( s == null ) {
          System.err.println("Symboltable returned from recursed tc is null");
        }
        s.isClass = true;
        s.type = classname;
        knownTypes.put( classname , s );
        genCode &= ltc.genCode;
        currentclass = "";
      }
    }
  }

  public void caseAFuncDecDeclaration( AFuncDecDeclaration node )
  {
    String name = node.getIdentifier().toString().trim();
    String type =  node.getType().toString().trim();

    ln = node.getIdentifier().getLine();

    if ( symbolTable.containsKey( name ) && ! ((SymbolTable)symbolTable.get( name )).prototypeOnly ) {

      giveError("A function with the name " + name + " has already been defined." , ln );
      return;

    } else {

      if (! knownTypes.containsKey( type ) ) {
        giveError("The return type " + type + " of function " + name + " is unknown." , ln );

      } else {

        SymbolTable entry = new SymbolTable( type );

        entry.isFunction = true;

        if ( node.getBracketPair().size() > 0 ) {
          entry.arraydim = node.getBracketPair().size();
        }

        symbolTable.put( name , entry );
        functions.put( name, entry );

        TypeChecker forfunc = new TypeChecker( ui , knownTypes , (SymbolTable)functions.clone() );
        forfunc.nowindowmode = nowindowmode;

        forfunc.currentfunc = name;

        if ( node.getFormalParamList() != null ) {
          node.getFormalParamList().apply( forfunc );
        }

        if ( node.getBlock() != null ) {
          node.getBlock().apply( forfunc );
        }

        SymbolTable lst = forfunc.getSymbolTable();
        lst.type = type;
        lst.isFunction = true;

        if ( node.getBracketPair().size() > 0 ) {
          lst.arraydim = node.getBracketPair().size();
        }

        symbolTable.put( name , lst );
        functions.put( name , lst );
        genCode &= forfunc.genCode;

        currentfunc = "";
      }
    }
  }

  public void caseAReturnStatement( AReturnStatement node )
  {
    ln = node.getReturn().getLine();

    if ( currentfunc.equals("") ) {
      giveError("return found outside a function" , ln );
      return;
    }

    String type = ((SymbolTable)symbolTable.get( currentfunc )).type;

    if ( type.equals("void") ) {
      if ( node.getExpression() != null )
        giveError("Value returned from void function " + currentfunc , ln );

      return;
    }

    if ( node.getExpression() == null && type.equals("void") ) {
      checked &= true;
      return;
    }

    if ( node.getExpression() == null ) {
      checked = false;
      return;
    } else {
      expected = type;
      node.getExpression().apply(this);
    }

    if ( ! expected.equals( found ) ) {
      giveError("Value returned from function " + currentfunc + " should have type " + type , ln );
    }
  }

  public void caseATypeName( ATypeName node )
  {
    String name = node.getIdentifier().toString().trim();
    String type = node.getType().toString().trim();
    ln = node.getIdentifier().getLine();

    if ( node.getBracketPair() != null ) {
      arraydim = node.getBracketPair().size();
    }

    if (! knownTypes.containsKey( type ) ) {
      giveError("The type " + type + " for function parameter " + name + " is unknown." , ln );

    } else {

      SymbolTable st;

      if ( arraydim > 0 ) {
        st  = (SymbolTable)((SymbolTable)knownTypes.get( type )).clone();
        st.arraydim = arraydim;
        st.isClass = true;
        st.put( "length" , (SymbolTable)knownTypes.get( "int" ) );
      } else {
        st = (SymbolTable)knownTypes.get( type );
      }

      symbolTable.put( name , st );
    }
  }

  public void caseAConstDecDeclaration( AConstDecDeclaration node )
  {
    String id = node.getIdentifier().toString().trim();
    String type =  node.getType().toString().trim();
    ln = node.getConst().getLine();

    if ( symbolTable.containsKey( id ) ) {
      giveError("A variable with the name " + id + " has already been declared." , ln );
      return;
    } else {

      if (! knownTypes.containsKey( type ) ) {
        giveError("The type " + type + " has not been defined." , ln );

      } else {

        if ( node.getInitialiser() != null ) {
          expected = type;
          node.getInitialiser().apply( this );
          if ( !checked ) {
            giveError( "Constant " + id + " must be initialised with type " + type , ln );
            return;
          }

        }

        SymbolTable st = (SymbolTable)((SymbolTable)knownTypes.get( type )).clone();
        st.constant = true;
        symbolTable.put( id , st );
      }
    }
  }

  public void caseAVarDecDeclaration( AVarDecDeclaration node )
  {
    String type =  node.getType().toString();
    ln = node.getIdentifier().getLine();

    type = type.trim();
    arraydim = 0;

    if ( !knownTypes.containsKey( type ) ) {
      giveError( "The type " + type + " has not been defined." , ln );
      return;
    }

    String id = node.getIdentifier().toString().trim();

    if ( symbolTable.containsKey( id ) ) {
      giveError("A variable with the name " + id + " has already been declared." , ln );
      return;
    }

    if ( node.getTypeParam() != null ) {
      node.getTypeParam().apply(this);
    }

    Node init = node.getInitialiser();

    if ( init != null )
      expectingArrayIndex = false;

    if ( node.getBracketPair().size() > 0 ) {

      if ( node.getArrayAccess().size() > 0 ) {
        giveError("Can't use [ ] and " + node.getArrayAccess().toString() + " in the same declaration." , ln );
        return;
      }

      if ( init == null ) {
        giveError("Array declared without size must be initialised." , ln );
        return;
      }

      arraydim = node.getBracketPair().size();
    }

    else if ( node.getArrayAccess() != null ) {

      LinkedList aas = node.getArrayAccess();
      ListIterator aaIterator = aas.listIterator(0);

      while ( aaIterator.hasNext() ) {
        arraydim++;
        Node t = (Node)aaIterator.next();
        t.apply( this );
      }

      expectingArrayIndex = true;
    }

    if ( init != null ) {
      expected = type;
      init.apply( this );
      if ( !checked ) {
        giveError( "Variable " + id + " must be initialised with type " + type , ln );
        return;
      }

      if ( maxdepth != arraydim &&  !( init instanceof AScalarInitialiser ) ) {
        giveError( "Array declared with " + arraydim + " dimensions. Initialiser has " + maxdepth + " dimensions." , ln );
      }

      maxdepth = curdepth = 0;
    }

    if ( arraydim == 0 ) {

      if ( type.equals( currentclass )) {

        symbolTable.put( id , symbolTable );

      } else {

        symbolTable.put( id , (SymbolTable)knownTypes.get( type ) );
      }  

    } else {

      if ( type.equals( currentclass )) {
        SymbolTable st = (SymbolTable)symbolTable.clone();
        st.arraydim = arraydim;
        st.isClass = true;
        st.put( "length" , (SymbolTable)knownTypes.get( "int" ) );
        symbolTable.put( id , st );

      } else {
        SymbolTable st = (SymbolTable)((SymbolTable)knownTypes.get( type )).clone(); 
        st.arraydim = arraydim;
        st.isClass = true;
        st.put( "length" , (SymbolTable)knownTypes.get( "int" ) );
        symbolTable.put( id , st );
      }
    }

    arraydim = 0;
  }

  public void caseATypeParamList( ATypeParamList node ) {

    if ( ! knownTypes.containsKey( node.getType().toString().trim() ) ) {
      giveError("Unknown type given as type parameter" , ln );
    }

    if ( node.getCommaType() != null ) {
      LinkedList ctl = node.getCommaType();
      ListIterator ctlIterator = ctl.listIterator(0);
      while ( ctlIterator.hasNext() ) {
        ACommaType ct = (ACommaType)ctlIterator.next();

        if ( ! knownTypes.containsKey( ct.getType().toString().trim() ) ) {
          giveError("Unknown type given as type parameter" , ln );
        }
      }
    }
  }

  public void caseAArrayInit( AArrayInit node ) {

    if ( ++curdepth > maxdepth )
      maxdepth = curdepth;

    if ( node.getInitList() != null ) {
      node.getInitList().apply( this );
    }

    --curdepth;
  }        

  public void caseAScalarInitialiser( AScalarInitialiser node )
  {
    if ( arraydim > 0 ) {
      node.getExpression().apply( this );
      if ( founddim == 0 ) {
        giveError( "Array cannot be initialised with a scalar value." , ln );
      }
    } else {
      node.getExpression().apply ( this );
    }
  }

  public StringPair separateOnDot( String from )
  {
    int p = from.indexOf('.');
    StringPair pair = new StringPair();

    if ( p < 0 ) {
      pair.a = pair.b = from;
      pair.separated = false;
    } else {
      pair.a = from.substring( 0 , p ).trim();
      pair.b = from.substring( p+1 , from.length() ).trim();
      pair.separated = true;
    }

    return pair;
  }

  private String nameFromFA( PFieldAccess node ) {
    if ( node instanceof AScalarFieldAccess ) {
      return((AScalarFieldAccess)node).getName().toString().trim();
    } else {
      return((AArrayFieldAccess)node).getName().toString().trim();
    }
  } 

  private String uptolbracket( String a ) {

    if ( a.indexOf('[') < 0 ) {
      return a.trim();
    } else {
      return a.substring( 0 , a.indexOf('[') ).trim();
    }
  }


  //***********For loops*****************

  public void caseAJavaForControl( AJavaForControl node ) {

    ln = node.getS1().getLine();

    if ( node.getExpression() != null ) {
      expected = "double";
      node.getExpression().apply(this);
      if ( checked == false ) {
        giveError("For loop starting index must be numeric" , ln ); return;
      }
    }

    if ( node.getBoolExpression() != null ) {
      expected = "boolean";
      node.getBoolExpression().apply(this);
      if ( checked == false ) {
        giveError("For loop termination condition must be boolean" , ln ); return;
      }
    }
  }

  public void caseASimpleForControl( ASimpleForControl node ) {

    ln = node.getTo().getLine();

    if ( node.getExp1() != null ) {
      expected = "double";
      node.getExp1().apply(this);
      if ( checked == false ) {
        giveError("For loop starting index must be numeric" , ln ); return;
      }
    }
    if ( node.getExp2() != null ) {
      expected = "double";
      node.getExp2().apply(this);
      if ( checked == false ) {
        giveError("For loop ending index must be numeric" , ln ); return;
      }
    }

    if ( node.getStepBit() != null ) {
      AStepBit sb = (AStepBit)node.getStepBit();
      expected = "double";
      sb.getExpression().apply(this);
      if ( checked == false ) {
        giveError("For loop step must be numeric" , ln ); return;
      }
    }
  }

  public void caseAWhileStatement( AWhileStatement node )
  {
    ln = node.getWhile().getLine();

    if ( node.getBoolExpression() != null ) {
      expected = "boolean";
      node.getBoolExpression().apply( this );

      if ( checked == false ) {
        giveError("While condition must be boolean" , ln );
        return;
      }
    }

    if ( node.getBlock() != null ) {

      expected = "";
      node.getBlock().apply( this );
    }
  }

  public void caseASwitchStatement( ASwitchStatement node ) {

    if ( node.getExpression() != null ) {

      expected = "int";
      node.getExpression().apply(this);

      if ( checked == false ) {

        checked = true;
        expected = "char";
        node.getExpression().apply( this );

        if ( checked == false ) {
          ln = node.getSwitch().getLine();
          giveError("You can only switch on an integer or a char." , ln );
          return;
        }
      }

      expected = "";

      if ( node.getSwitchBlock() != null ) {
        node.getSwitchBlock().apply( this );
      }
    }
  }

  public void caseAAssignmentStatement( AAssignmentStatement node )
  {
    ln = node.getAssign().getLine();

    PFieldAccess fieldaccess = node.getFieldAccess();

    fieldaccess.apply(this);

    String name = nameFromFA( fieldaccess );
    SymbolTable s = symbolTable;
    StringPair ab = separateOnDot( name );


    while ( ab.separated ) {
      String arrname = uptolbracket( ab.a );

      if ( ! s.containsKey( arrname ) ) {
        giveError("Variable " + arrname + " has not been declared" , ln );
        return;

      } else {

        s = ((SymbolTable)symbolTable.get( arrname ));
        if ( ! s.isClass ) {
          giveError( ab.a + " is not a class." , ln );
          return;
        }
      }

      ab = separateOnDot( ab.b ); 
    }

    if ( ! s.containsKey( ab.b ) ) {
      giveError("Variable " + name + " has not been declared" , ln );
    } else {

      try {

        SymbolTable st = (SymbolTable)s.get( ab.b );

        if ( st.constant ) {
          giveError( "Cannot assign to constant " + ab.a , ln );
          return;
        }

        if ( st.isFunction ) {
          giveError( ab.a + " is a function and cannot be assigned to." , ln );
          return;
        }

        expected = st.type;
        expdim = st.arraydim;

        if ( expected == null ) System.err.println("expected is null in assignment");

        if ( node.getExpression() != null ) {
          node.getExpression().apply( this );
        }


      } catch ( NullPointerException e ) {
        e.printStackTrace();
      }
    }

    if ( ! checked )
      giveError( "Expression assigned to " + name + " should have type " + expected , ln );

  }

  public void caseAPrintStrStatement( APrintStrStatement node ) {

    ln = node.getPrint().getLine();

    if ( node.getExpression() != null ) {

      expected = "";
      node.getExpression().apply( this );
    }
  }

  public void caseAPrintlnStrStatement( APrintlnStrStatement node ) {

    ln = node.getPrintln().getLine();

    if ( node.getExpression() != null ) {
      expected = "";
      node.getExpression().apply( this );
    }
  }


  // ************* Graphics Primitives ***************** //

  public void caseADrawstringStatement( ADrawstringStatement node ) {

    ln = node.getDrawstring().getLine();

    expected = "String";
    node.getString().apply( this );
    if ( !checked ) {
      giveError("First argument to drawString must be a String" , ln );
      return;
    }

    expected = "int";
    node.getX().apply( this );
    if ( !checked ) {
      giveError("Second argument to drawString (x co-ord) must be an int" , ln );
      return;
    }

    expected = "int";
    node.getY().apply( this );
    if ( !checked ) {
      giveError("Third argument to drawString (y co-ord) must be an int" , ln );
      return;
    }
  }

  public void caseAFillStatement( AFillStatement node ) {

    ln = node.getFill().getLine();
    expected = "boolean";
    node.getBoolExpression().apply( this );
    if ( ! checked ) {
      giveError("Fill takes a boolean parameter." , ln );
    }

  }

  public void caseAFourNumbers( AFourNumbers node ) {

    ln = node.getC1().getLine();

    if ( node.getX1() != null ) {

      expected = "int";
      node.getX1().apply( this );

      if ( !checked ) {
        giveError("First parameter should have type int" , ln );
      }
    }

    if ( node.getY1() != null ) {

      expected = "int";
      node.getY1().apply( this );

      if ( !checked ) {
        giveError("Second parameter should have type int" , ln );
      }
    }

    if ( node.getX2() != null ) {

      expected = "int";
      node.getX2().apply( this );

      if ( !checked ) {
        giveError("Third parameter should have type int" , ln );
      }
    }

    if ( node.getY2() != null ) {

      expected = "int";
      node.getY2().apply( this );

      if ( !checked ) {
        giveError("Fourth parameter should have type int" , ln );
      }
    }
  }

  public void caseASixNumbers ( ASixNumbers node ) {

    if ( node.getFourNumbers() != null ) {
      node.getFourNumbers().apply( this );
    }

    if ( node.getX() != null ) {

      expected = "int";
      node.getX().apply( this );

      if ( !checked ) {
        giveError("Fifth parameter should have type int" , ln );
      }
    }

    if ( node.getY() != null ) {

      expected = "int";
      node.getY().apply( this );

      if ( !checked ) {
        giveError("Sixth parameter should have type int" , ln );
      }
    }
  }

  public void caseARgb( ARgb node ) {

    ln = node.getC1().getLine();

    if ( node.getR() != null ) {

      expected = "int";
      node.getR().apply( this );

      if ( !checked ) {
        giveError("First (red) parameter should have type int" , ln );
      }
    }

    if ( node.getG() != null ) {

      expected = "int";
      node.getG().apply( this );

      if ( !checked ) {
        giveError("Second (green) parameter should have type int" , ln );
      }
    }

    if ( node.getB() != null ) {

      expected = "int";
      node.getB().apply( this );

      if ( !checked ) {
        giveError("Third (blue) parameter should have type int" , ln );
      }
    }
  }


  //********** Expression terms and factors ******************//


  public void caseAPowerTerm( APowerTerm node )
  {
    checked &= ( expected.equals("double") || expected.equals("") );
    if ( ! checked ) {
      return;
    }
    expected = "double";

    if ( node.getTerm() != null ) {
      node.getTerm().apply(this);
    }

    if ( node.getUnaryExp() != null ) {
      node.getUnaryExp().apply(this);
    }

    found = "double";
  }

  public void caseAFunctionApplication( AFunctionApplication node )
  {
    String name = node.getName().toString().trim();
    ln = node.getLParenthese().getLine();

    if ( !symbolTable.containsKey( name ) ) {
      giveError( "Function " + name + " has not been defined." , ln );
      return;
    }

    SymbolTable s = (SymbolTable)symbolTable.get( name );
    String type = s.type;
    //System.out.println("\n\nFunction : "+name);
    //System.out.println("Expected  : "+expected);

    if ( expected.equals( "double" ) ) {
      checked &= ( type.equals("int") || type.equals("double") );
    } else if ( expected.equals( "int" ) ) {
      checked &= ( type.equals( "int" ) || type.equals("char") );
    } else {
      checked &= ( expected.equals( type ) || expected.equals("") );    
    }

    //System.out.println("Checked : "+checked);
    //System.out.println("Type : "+type);

    found = type;
    founddim = s.arraydim;
  }

  public void caseAFNumber( AFNumber node ) {

    checked &= ( expected.equals( "double" ) || expected.equals("") );
    found = "double";
  }

  public void caseAINumber( AINumber node ) {

    checked &= ( expected.equals( "int" ) || expected.equals( "double") || expected.equals("") );
    found = "int"; 
  }

  public void caseAStringliteralFactor( AStringliteralFactor node ) {

    checked &= ( expected.equals( "String" )  || expected.equals("") );
    found = "String";
  }

  public void caseACharliteralFactor( ACharliteralFactor node ) {

    checked &= ( expected.equals( "char" ) || expected.equals("") );
    found = "char";
  }

  public void caseABoolliteralFactor( ABoolliteralFactor node ) {

    checked &= ( expected.equals( "boolean" ) || expected.equals("") );
    found = "boolean";
  }

  public void caseAArrayAccess( AArrayAccess node ) {

    ln = node.getLBracket().getLine();

    if ( expectingArrayIndex && node.getMathExpression() == null ) {
      giveError("Missing array index" , ln );
      return;
    }

    if ( ! expectingArrayIndex && node.getMathExpression() != null ) {
      giveError("Not expecting array index." , ln );
      return;
    }

    if ( ! expectingArrayIndex ) return;

    if ( node.getMathExpression() != null ) {
      expected = "int";
      node.getMathExpression().apply(this);
      if ( !checked ) {
        giveError("Array index must be integer" , ln = node.getLBracket().getLine() );
      }
    } else {
      giveError("Missing array index" , ln = node.getLBracket().getLine() );
    }
  }

  public void caseAReadFactor( AReadFactor node ) {

    checked &= ( expected.equals( "char" ) || expected.equals( "int" ) || expected.equals("") );
    found = "char";
  }

  public void caseAReadintFactor( AReadintFactor node ) {

    checked &= ( expected.equals( "int" )  || expected.equals("") );
    found = "int";
  }

  public void caseAReadrealFactor( AReadrealFactor node ) {

    checked &= ( expected.equals( "double" ) || expected.equals("") );
    found = "double";
  }

  public void caseAReadstringFactor( AReadstringFactor node ) {

    checked &= ( expected.equals( "String" ) || expected.equals("") );
    found = "String";
  }

  public void caseAReadcharFactor( AReadcharFactor node ) {

    checked &= ( expected.equals( "char" ) || expected.equals( "int" ) || expected.equals("") );
    found = "char";
  }

  public void caseASinFactor( ASinFactor node ) {

    ln = node.getSin().getLine();
    checked &= ( expected.equals( "double") || expected.equals("") );
    if ( !checked ) {
      return;
    }
    expected = "double";
    node.getMathExpression().apply( this );
    if ( !checked ) {
      giveError( "Argument to sin must be numeric" , ln );
    }
    found = "double";
  }

  public void caseAAsinFactor( AAsinFactor node ) {

    ln = node.getAsin().getLine();
    checked &= ( expected.equals( "double") || expected.equals("") );
    if ( !checked ) {
      return;
    }
    expected = "double";
    node.getMathExpression().apply( this );
    if ( !checked ) {
      giveError( "Argument to asin must be numeric" , ln );
    }
    found = "double";
  }

  public void caseACosFactor( ACosFactor node ) {
    ln = node.getCos().getLine();
    checked &= ( expected.equals( "double") || expected.equals("") );
    if ( !checked ) {
      return;
    }
    expected = "double";
    node.getMathExpression().apply( this );
    if ( !checked ) {
      giveError( "Argument to cos must be numeric" , ln );
    }
    found = "double";
  }

  public void caseAAcosFactor( AAcosFactor node ) {

    checked &= ( expected.equals( "double") || expected.equals("") );
    if ( !checked ) {
      return;
    }
    expected = "double";
    node.getMathExpression().apply( this );
    if ( !checked ) {
      giveError( "Argument to acos must be numeric" , node.getAcos().getLine() );
    }
    found = "double";
  }

  public void caseATanFactor( ATanFactor node ) {

    ln = node.getTan().getLine();
    checked &= ( expected.equals( "double") || expected.equals("") );
    if ( !checked ) {
      return;
    }
    expected = "double";
    node.getMathExpression().apply( this );
    if ( !checked ) {
      giveError( "Argument to tan must be numeric" , ln );
    }
    found = "double";
  }

  public void caseAAtanFactor( AAtanFactor node ) {

    ln = node.getAtan().getLine();
    checked &= ( expected.equals( "double") || expected.equals("") );
    if ( !checked ) {
      return;
    }
    expected = "double";
    node.getMathExpression().apply( this );
    if ( !checked ) {
      giveError( "Argument to atan must be numeric" , ln );
    }
    found = "double";
  }

  public void caseASqrtFactor( ASqrtFactor node ) {

    ln =  node.getSqrt().getLine();
    checked &= ( expected.equals( "double") || expected.equals("") );
    if ( !checked ) {
      return;
    }
    expected = "double";
    node.getMathExpression().apply( this );
    if ( !checked ) {
      giveError( "Argument to sqrt must be numeric" , ln );
    }
    found = "double";
  }

  public void caseALogFactor( ALogFactor node ) {

    ln =  node.getLog().getLine();
    checked &= ( expected.equals( "double") || expected.equals("") );
    if ( !checked ) {
      return;
    }
    expected = "double";
    node.getMathExpression().apply( this );
    if ( !checked ) {
      giveError( "Argument to log must be numeric" , ln );
    }
    found = "double";
  }

  public void caseAExpFactor( AExpFactor node ) {

    ln =  node.getExp().getLine();
    checked &= ( expected.equals( "double") || expected.equals("") );
    if ( !checked ) {
      return;
    }
    expected = "double";
    node.getMathExpression().apply( this );
    if ( !checked ) {
      giveError( "Argument to exp must be numeric" , ln );
    }
    found = "double";
  }    

  public void caseARoundFactor( ARoundFactor node ) {

    ln = node.getRound().getLine();
    checked &= ( expected.equals( "int" ) || expected.equals("") );
    if ( !checked ) {
      return;
    }
    expected = "double";
    node.getMathExpression().apply( this );
    if ( !checked ) {
      giveError( "Argument to round must be numeric" , ln );
    }
    found = "int";
  }    

  public void caseAFloorFactor( AFloorFactor node ) {

    ln = node.getFloor().getLine();
    checked &= ( expected.equals( "int" ) || expected.equals( "double" ) || expected.equals("") );
    if ( !checked ) {
      return;
    }
    expected = "double";
    node.getMathExpression().apply( this );
    if ( !checked ) {
      giveError( "Argument to floor must be numeric" , ln );
    }
    found = "int";
  }

  public void caseARandomFactor( ARandomFactor node ) {

    checked &= ( expected.equals( "double" ) || expected.equals("") ); 
    found = "double";
  }

  public void caseACeilFactor( ACeilFactor node ) {

    ln =  node.getCeil().getLine();
    checked &= ( expected.equals( "int" ) || expected.equals( "double" ) || expected.equals("") );
    if ( !checked ) {
      return;
    }
    expected = "double";
    node.getMathExpression().apply( this );
    if ( !checked ) {
      giveError( "Argument to ceil must be numeric" , ln );
    }
    found = "double";
  }    

  public void caseAFieldaccessFactor( AFieldaccessFactor node ) {

    PFieldAccess fieldaccess = node.getFieldAccess();

    String name = nameFromFA( fieldaccess );
    SymbolTable s = symbolTable;
    StringPair ab = separateOnDot( name );
    String arrname = uptolbracket( ab.a );

    while ( ab.separated ) {
      if ( ! s.containsKey( arrname ) ) {
        giveError("Variable " + arrname + " has not been declared" , ln );
        return;

      } else {
        s = ((SymbolTable)s.get( arrname ));
        if ( ! s.isClass ) {
          giveError( ab.a + " is not a class." , ln );
          return;
        }
      }

      ab = separateOnDot( ab.b );
      arrname = uptolbracket( ab.a );
    }

    if ( ! s.containsKey( ab.b ) ) {

      giveError("Variable " + ab.b + " has not been declared." , ln );

    } else {

      String type = ((SymbolTable)s.get( ab.b )).type;

      if ( type != null ) {
        if ( expected.equals( "double" ) ) {

          checked &= ( type.equals("int") || type.equals("double") );

        } else if ( expected.equals("int") ) {

          checked &= ( type.equals("int") || type.equals("char") );

        } else {
          checked &= ( expected.equals( type ) || expected.equals("") );
        }

        found = type;

      } else {
        System.err.println("type in symbol table for " + name + " is null");
      }
    }
  }

  public void caseANullFactor( ANullFactor node ) {

    //matches any type

    found = expected;
  }

  /***** relational expressions *****/

  public void caseALtRelational( ALtRelational node ) {

    checked &= expected.equals("boolean");
    if ( !checked ) {
      found = "boolean";
      return;
    }

    expected = "";

    ln = node.getLess().getLine();

    if ( node.getRelational() != null ) {
      node.getRelational().apply( this );
    }

    String lhs_type = expected = found;

    if ( node.getMathExpression() != null ) {
      node.getMathExpression().apply( this );
    }

    if ( !checked ) {
      giveError( "Cannot compare " + lhs_type + " to " + found + " with < " , ln );
    }
  }

  public void caseAGtRelational( AGtRelational node ) {

    checked &= expected.equals("boolean");
    if ( !checked ) {
      found = "boolean";
      return;
    }

    expected = "";

    ln = node.getGreater().getLine();

    if ( node.getRelational() != null ) {
      node.getRelational().apply( this );
    }

    String lhs_type = expected = found;

    if ( node.getMathExpression() != null ) {
      node.getMathExpression().apply( this );
    }

    if ( !checked ) {
      giveError( "Cannot compare " + lhs_type + " to " + found + " with > " , ln );
    }
  }

  public void caseALteqRelational( ALteqRelational node ) {

    checked &= expected.equals("boolean");
    if ( !checked ) {
      found = "boolean";
      return;
    }

    expected = "";

    ln = node.getLessequal().getLine();

    if ( node.getRelational() != null ) {
      node.getRelational().apply( this );
    }

    String lhs_type = expected = found;

    if ( node.getMathExpression() != null ) {
      node.getMathExpression().apply( this );
    }

    if ( !checked ) {
      giveError( "Cannot compare " + lhs_type + " to " + found + " with <= " , ln );
    }
  }

  public void caseAGteqRelational( AGteqRelational node ) {

    checked &= expected.equals("boolean");
    if ( !checked ) {
      found = "boolean";
      return;
    }

    expected = "";

    ln = node.getGreaterequal().getLine();

    if ( node.getRelational() != null ) {
      node.getRelational().apply( this );
    }

    String lhs_type = expected = found;

    if ( node.getMathExpression() != null ) {
      node.getMathExpression().apply( this );
    }

    if ( !checked ) {
      giveError( "Cannot compare " + lhs_type + " to " + found + " with >= " , ln );
    }
  }

  public void caseAEqEquality( AEqEquality node ) {

    checked &= expected.equals("boolean");
    if ( !checked ) {
      found = "boolean";
      return;
    }

    expected = "";

    ln = node.getEqual().getLine();

    if ( node.getEquality() != null ) {
      node.getEquality().apply( this );
    }

    String lhs_type = expected = found;

    if ( node.getRelational() != null ) {
      node.getRelational().apply( this );
    }

    if ( !checked ) {
      giveError( "Cannot compare " + lhs_type + " to " + found + " with == " , ln );
    }
  }

  public void caseANeqEquality( ANeqEquality node ) {

    checked &= expected.equals("boolean");
    if ( !checked ) {
      found = "boolean";
      return;
    }

    expected = "";

    ln = node.getNotequal().getLine();

    if ( node.getEquality() != null ) {
      node.getEquality().apply( this );
    }

    String lhs_type = expected = found;

    if ( node.getRelational() != null ) {
      node.getRelational().apply( this );
    }

    if ( !checked ) {
      giveError( "Cannot compare " + lhs_type + " to " + found + " with != " , ln );
    }
  }

  public void caseAAndBoolTerm( AAndBoolTerm node )
  {
    expected = "boolean";

    if ( node.getBoolTerm() != null )
      node.getBoolTerm().apply( this );

    if ( !checked )
      giveError( found + " expression found where boolean expected, before " + node.getAnd().toString() , ln );

    expected = "boolean";

    if ( node.getEquality() != null )
      node.getEquality().apply( this );

    if ( !checked )
      giveError( found + " expression found where boolean expected, after " + node.getAnd().toString() , ln );
  }

  public void caseAOrBoolExpression( AOrBoolExpression node )
  {
    expected = "boolean";

    if ( node.getBoolExpression() != null )
      node.getBoolExpression().apply( this );

    if ( !checked )
      giveError( found + " expression found where boolean expected, after " + node.getOr().toString() , ln );

    expected = "boolean";

    if ( node.getBoolTerm() != null )
      node.getBoolTerm().apply( this );

    if ( !checked )
      giveError( found + " expression found where boolean expected, after " + node.getOr().toString() , ln );
  }

  public void caseAXorBoolExpression( AXorBoolExpression node )
  {
    expected = "boolean";

    if ( node.getBoolExpression() != null )
      node.getBoolExpression().apply( this );

    if ( !checked )
      giveError( found + " expression found where boolean expected, after " + node.getXor().toString() , ln );

    expected = "boolean";

    if ( node.getBoolTerm() != null )
      node.getBoolTerm().apply( this );

    if ( !checked )
      giveError( found + " expression found where boolean expected, after " + node.getXor().toString() , ln );
  }

  /******************* Other statments ******8************/

  public void caseAAssertStatement( AAssertStatement node )
  {
    ln = node.getAssert().getLine();

    if ( node.getBoolExpression() != null ) {
      expected = "boolean";
      node.getBoolExpression().apply( this );
      if ( ! checked ) {
        giveError( "Can only assert boolean conditions" , ln );
        return;
      }
    }

    if ( node.getColonString() != null ) {
      node.getColonString().apply( this );
    }
  }

  public void caseAColonString( AColonString node )
  {
    expected = "String";

    if ( node.getExpression() != null )
      node.getExpression().apply( this );

    if ( !checked )
      giveError( "Assertion message must be a String." , ln );
  }

  public void caseAArraycopyStatement( AArraycopyStatement node )
  {
    expected = "";
    node.getSrc().apply( this );

    expected = found;

    node.getDest().apply( this );

    if ( !checked ) {
      giveError( "arraycopy can only be applied to two arrays of the same type" );
    }
  }

  public void caseAIfStatement( AIfStatement node )
  {
    ln = node.getIf().getLine();

    if ( node.getBoolExpression() != null ) {
      expected = "boolean";
      node.getBoolExpression().apply( this );
      if ( !checked ) {
        giveError("Condition must be boolean" , ln );
        return;
      }
    }

    if ( node.getBlock1() != null )
      node.getBlock1().apply( this );

    if ( node.getBlock2() != null )
      node.getBlock2().apply( this );
  }

  public void giveError( String s , int line )
  {
    if ( nowindowmode ) {
      System.err.println( "Error (line " + line + "): " + s + "\n" );
    } else {
      ui.msgsArea.append( "Error (line " + line + "): " + s + "\n" );

      try {

        java.awt.Point start = new java.awt.Point( 0 , line * ui.fontheight );
        java.awt.Point end = new java.awt.Point( ui.kenyaArea.getWidth() , line * ui.fontheight  );
        ui.highlight.addHighlight( ui.kenyaArea.viewToModel( start ) , ui.kenyaArea.viewToModel( end ) , ui.painter );

      } catch ( javax.swing.text.BadLocationException e ) {
        System.err.println(e);
      }
    }
    genCode = false;
    checked = true;
  }

  public void giveError( String s )
  {
    if ( nowindowmode ) {
      System.err.println( "Error : " + s + "\n" );
    } else {
      ui.msgsArea.append( "Error : " + s + "\n" );
    }
    genCode = false;
    checked = true;
  }

}
